#include "qam.h"

#include <math.h>

bool qam_prepare_config(qam_config_t *config) {
  config->valid = false;
  if (!(
    config->sample_rate > 0 &&
    config->carrier_freq > 0 &&
    config->sample_rate % config->carrier_freq == 0 &&
    config->switch_half_phases >= 0 && config->switch_half_phases < 256 &&
    config->hold_half_phases >= 2 && config->hold_half_phases < 256 &&
    (
      (
        (config->switch_half_phases & 1) == 0 &&
        (config->hold_half_phases & 1) == 0
      ) || ((config->sample_rate / config->carrier_freq) & 1) == 0
    ) &&
    config->bits_per_symbol >= 2 &&
    (config->bits_per_symbol & 1) == 0 &&
    (config->qamsync_symbol_count & 1) != 0 &&
    config->qamsync_resync_interval >= 2 &&
    config->symsync_resync_threshold >= 1 &&
    config->tx_delay >= 0.0f &&
    config->symsync_tx_rx_ratio >= 1.0f &&
    config->symsync_tolerance > 0.0f &&
    config->qamsync_tolerance > 0.0f
  )) return false;
  config->max_amplitude_factor = (1 << (config->bits_per_symbol / 2)) - 1;
  {
    double precision = 0.5 / config->max_amplitude_factor * M_PI;
    config->symsync_phase_precision = (float)(
      precision * config->symsync_tolerance
    );
    if (config->symsync_phase_precision >= 0.25 * M_PI) {
      config->symsync_phase_precision = (float)(0.25 * M_PI);
    }
    config->qamsync_phase_precision = (float)(
      precision * config->qamsync_tolerance
    );
  }
  config->tx_amplitude = (float)(
    (double)config->tx_volume / M_SQRT2 / config->max_amplitude_factor
  );
  config->switch_odd = config->switch_half_phases & 1 ? -1 : 1;
  config->hold_odd = config->hold_half_phases & 1 ? -1 : 1;
  {
    const int phase_samples = config->sample_rate / config->carrier_freq;
    config->switch_samples = (phase_samples * config->switch_half_phases) / 2;
    config->hold_samples = (phase_samples * config->hold_half_phases) / 2;
    config->symbol_samples = config->switch_samples + config->hold_samples;
    config->init_samples = config->switch_samples + 2 * config->hold_samples;
    config->carrier_phase_step = (float)(M_PI / (phase_samples / 2.0));
    config->switch_phase_step = (float)(M_PI / config->switch_samples);
  }
  config->symsync_rx_count = (config->symbol_samples + 1) / 2;
  if (config->symsync_rx_count < 25) config->symsync_rx_count = 25;
  config->symsync_tx_half_count = round(
    (
      (double)config->symsync_rx_count * config->symsync_tx_rx_ratio +
      (double)config->tx_delay * config->sample_rate / config->symbol_samples
    ) / 2.0
  );
  if (!(
    config->qamsync_symbol_count * config->hold_samples >= config->init_samples
  )) return false;
  config->valid = true;
  return true;
}

